import functools
import time
import gzip
import os
import pysam
import re
import sys


def log_info(func):
    """Decorator that prints function arguments and runtime
    """

    @functools.wraps(func)
    def wrapper(args):
        print(
            "Function {} called with the following arguments:\n".format(func.__name__),
            file=sys.stderr
        )
        for arg in vars(args):
            print(str(arg) + "\t" + str(getattr(args, arg)), file=sys.stderr)
        t1 = time.time()
        func(args)
        t2 = time.time()
        elapsed = [round(x, 2) for x in divmod(t2 - t1, 60)]
        print("\nFunction completed in  {} m {} s\n".format(elapsed[0], elapsed[1]),
             file=sys.stderr)

    return wrapper


def chunk_bam(bamfile, nproc):
    """
    chunk file into n chunks for multicore
    """
    chrom_lengths = bamfile.lengths
    chunksize = sum(chrom_lengths) / int(nproc)
    intervals = []
    for x in range(1, nproc + 1):
        position = chunksize * x
        intervals.append(find_chromosome_break(position, chrom_lengths, 0))
    return add_start_coords(intervals, chrom_lengths, bamfile)


def add_start_coords(intervals, chrom_lengths, bamfile):
    """
    given the intervals that will be handled by each core,
    break into genomic regions (removing chromosome-spanning intervals)
    """
    intervals = [[1, 0]] + intervals
    ranges = [intervals[x - 1] + intervals[x] for x in range(1, len(intervals))]
    d = {}
    x = 0
    for i in ranges:
        x += 1
        if i[0] == i[2]:
            d[x] = [(bamfile.get_reference_name(i[0] - 1), i[1], i[3])]
        else:
            d[x] = [
                (bamfile.get_reference_name(i[0] - 1), i[1], chrom_lengths[i[0] - 1])
            ]
            nchrom = i[2] - i[0]
            for y in range(nchrom - 1):
                d[x].append(
                    (bamfile.get_reference_name(i[0] + y), 0, chrom_lengths[i[0] + y])
                )
            d[x].append((bamfile.get_reference_name(i[2] - 1), 0, i[3]))
    return d


def find_chromosome_break(position, chromosomes, current_chrom):
    assert position <= sum(chromosomes), "position past end of genome"
    if position <= chromosomes[current_chrom]:
        return [current_chrom + 1, position]
    else:
        position = position - chromosomes[current_chrom]
        return find_chromosome_break(position, chromosomes, current_chrom + 1)


def chunk(seq, num):
    """
    cut list into n chunks
    """
    avg = len(seq) / float(num)
    out = []
    last = 0.0
    while last < len(seq):
        out.append(seq[int(last) : int(last + avg)])
        last += avg
    return out


def scan_tags(tags, cb="CB", ub="UB"):
    """
    Input bam tags
    Return UMI and cell barcode sequences

    Parameters
    ----------
    tags
        List of read tags
    cb : str
        Tag for cell barcode. Default is CB, as used by 10x
    ub : str
        Tag for UMI barcode. Default is UB, as used by 10x
    """
    cell_barcode = None
    umi = None
    for tag in tags:
        if tag[0] == cb:
            cell_barcode = tag[1]
        elif tag[0] == ub:
            umi = tag[1]
        else:
            pass
    return cell_barcode, umi


def read_cells(cells):
    """Read file containing cell barcodes"""
    if cells is None:
        return None
    if os.path.isfile(cells):
        if cells.endswith(".gz"):
            cb = [line.strip("\n") for line in gzip.open(cells, "b")]
        else:
            cb = [line.strip("\n") for line in open(cells, "r")]
    else:
        cb = cells.split(",")
    return cb


def get_chromosomes(bam, keep_contigs="(?i)^chr"):
    """Create one interval for each chromosome"""
    if keep_contigs is None:
        keep_contigs = "."
    pattern = re.compile(keep_contigs)
    aln = pysam.AlignmentFile(bam, 'rb')
    idxstats = aln.get_index_statistics()
    keep_contigs = []
    for i in idxstats:
        if i.mapped > 0 and pattern.match(i.contig):
            keep_contigs.append(i.contig)
    conlen = {x: aln.get_reference_length(x) for x in keep_contigs}
    aln.close()
    return conlen


def read_cell_barcode_tag_file(infile):
    """
    Read in table of cell barcodes and associated group

    Note that each cell barcode can be in multiple groups

    Returns a dictionary where the cell barcode is the key,
    value is a list of tuples. Each tuple in the list is the 
    read tag and the group.

    Parameters
    ----------
    infile : str
        File name. Can be a gzip-compressed file or plain text.
    """
    cb = {}
    if os.path.isfile(infile):
        if infile.endswith(".gz"):
            inf = gzip.open(infile, "b")
        else:
            inf = open(infile, "r")
        for line in inf:
            line = line.rsplit()
            if line[0] in cb.keys():
                cb[line[0]].append((line[1], line[2]))
            else:
                cb[line[0]] = [(line[1], line[2])]
        inf.close()
        return cb
    else:
        raise Exception("File not found")


def read_cell_barcode_file(infile):
    """
    Read in table of cell barcodes and associated group

    Note that each cell barcode can be in multiple groups

    Returns a dictionary where the cell barcode is the key,
    value is a list of groups that the cell belongs to.

    Parameters
    ----------
    infile : str
        File name. Can be a gzip-compressed file or plain text.
    """
    cb = {}
    if os.path.isfile(infile):
        if infile.endswith(".gz"):
            inf = gzip.open(infile, "b")
        else:
            inf = open(infile, "r")
        for line in inf:
            line = line.rsplit()
            groups = line[1].split(",")
            if line[0] in cb.keys():
                cb[line[0]] += groups
                cb[line[0]] = list(set(cb[line[0]]))
            else:
                cb[line[0]] = groups
        inf.close()
        return cb
    else:
        raise Exception("File not found")


import sys
from contextlib import closing

import pysam
from sinto.constants import OUT_FORMAT_CONVERSION


def header_line_to_str(line):
    return "\t".join(f"{k}:{v}" for k, v in line.items())


def build_header(header, tag_vals):
    new_rg_lines = []
    for val in tag_vals:
        for rg in header["RG"]:
            rg = rg.copy()
            rg["SM"] += f":{val}"
            rg["ID"] += f":{val}"
            new_rg_lines.append("@RG\t" + header_line_to_str(rg))
    return str(header) + "\n".join(new_rg_lines) + "\n"


def tagtorg(bam, tag, output, tag_value_file, out_format="t"):
    """Add tags to reads from individual cells

    Copies BAM entries to a new file, adding a read tag to cells matching an input table

    Parameters
    ----------
    bam : str
        Path to SAM/BAM file, or "-" to read from stdin.
    tag_value_file : str
        Text file with one line per expected tag value.
    output : str
        Name for output SAM/BAM file, or "-" to write to stdout.
    out_format : str
        One of "t" (SAM), "b" (BAM), or "u" (uncompressed BAM) ("t" default)
    """

    infile = pysam.AlignmentFile(bam)
    with open(tag_value_file) as fh:
        tag_vals = {val.rstrip() for val in fh.readlines()}
    new_header = build_header(infile.header, tag_vals)

    outfile = pysam.AlignmentFile(
        output, "w" + OUT_FORMAT_CONVERSION[out_format], text=new_header
    )
    with closing(outfile) as outfile:
        for rec in infile:
            try:
                tag_val = rec.get_tag(tag)
            except KeyError:
                pass
            else:
                if tag_val in tag_vals:
                    rec.set_tag(
                        "RG",
                        f"{rec.get_tag('RG')}:{tag_val}",
                        value_type="Z",
                        replace=True,
                    )
            finally:
                outfile.write(rec)



import gzip
import os


def addbarcodes(cb_position, fq1, fq2, fq3=None, prefix="", suffix=""):
    """Add cell barcode to read names

    Parameters
    ----------
    cb_position : int
        Base positions containing cell barcode information
    fq1 : str
        Fastq file containing cell barcode sequences
    fq2 : str
        Fastq file to add cell barcodes to
    fq3 : str
        Fastq file to add cell barcodes to (optional)
    prefix : str
        Prefix to append to cell barcodes
    suffix : str
        Suffix to append to cell barcodes
    """
    barcodes = get_barcodes(f=fq1, bases=cb_position, prefix=prefix, suffix=suffix)
    add_barcodes(f=fq2, cb=barcodes)
    if fq3 is not None:
        add_barcodes(f=fq3, cb=barcodes)


def get_barcodes(f, bases=12, prefix="", suffix=""):
    f_open = open_fastq(f)
    if f.endswith(".gz"):
        gz = True
    else:
        gz = False
    x = 0
    cb = []
    for i in f_open:
        if (x % 4 == 1):
            if gz:
                cb.append(prefix + i.decode("utf-8")[:bases] + suffix)
            else:
                cb.append(prefix + i[:bases] + suffix)
        x += 1
    f_open.close()
    return(cb)


def add_barcodes(f, cb):
    f_open = open_fastq(f)
    if f.endswith(".gz"):
        gz = True
        o = f.replace(".fastq.gz", "").replace(".fq.gz", "") + ".barcoded.fastq.gz"
        outfile = gzip.GzipFile(o, mode = "wb")
    else:
        gz = False
        o = f.replace(".fastq", "").replace(".fq", "") + ".barcoded.fastq"
        outfile = open(o, "w+")
    x = 0
    y = 0
    for i in f_open:
        if (x % 4 ==0):
            if gz:
                rdname = i.decode("utf-8")
            else:
                rdname = i
            rdname = "@" + cb[y] + ":" + rdname[1:]
            if gz:
                outfile.write(rdname.encode("utf-8"))
            else:
                outfile.write(rdname)
            y += 1
        else:
            outfile.write(i)
        x += 1
    f_open.close()
    outfile.close()


def open_fastq(f):
    if os.path.isfile(f):
        if f.endswith(".gz"):
            f_open = gzip.open(f, "rb")
        else:
            f_open = open(f, "r")
        return(f_open)
    else:
        raise Exception("File not found")



from contextlib import closing

import pysam

from sinto.constants import OUT_FORMAT_CONVERSION


def tagtotag(
    bam: str,
    output: str,
    from_tag: str,
    to_tag: str,
    delete: bool,
    out_format: str = "t",
) -> None:
    """Copy/move tags

    Copies BAM entries to a new file while copying a read tag to another read tag
    and optionally deleting the originating tag.

    Parameters
    ----------
    bam : str
        Path to SAM/BAM file, or "-" to read from stdin.
    output : str
        Name for output SAM/BAM file, or "-" to write to stdout.
    from_tag: str
        Read tag to copy from.
    to_tag: str
        Read tag to copy to
    delete: bool
        Delete from_tag from read if True.
    out_format : str
        One of "t" (SAM), "b" (BAM), or "u" (uncompressed BAM) ("t" default)
    """
    assert len(from_tag) == 2
    assert len(to_tag) == 2

    infile = pysam.AlignmentFile(bam)
    outfile = pysam.AlignmentFile(
        output, "w" + OUT_FORMAT_CONVERSION[out_format], template=infile
    )

    for rec in infile:
        try:
            tag_val, tag_type = rec.get_tag(from_tag, with_value_type=True)
        except KeyError:
            pass
        else:
            rec.set_tag(to_tag, tag_val, value_type=tag_type, replace=True)
            if delete:
                rec.set_tag(from_tag, None)
        finally:
            outfile.write(rec)



import pysam
import re
from sinto import utils
from sinto.constants import OUT_FORMAT_CONVERSION
from contextlib import closing


def move(
    bam: str,
    output: str,
    cb_tag: str = "CB",
    cb_regex: str = "[^:]*",
    from_tag: bool = True,
    out_format: str = "t",
) -> None:
    """Copy cell barcode from tag to read name or vice versa
    
    Copies BAM entries to a new file. Note that cell barcodes are copied, the original tag/readname is not removed.

    Parameters
    ----------
    bam : str
        Path to BAM file.
    output : str
        Name of output BAM or SAM file
    cb_tag : str
        Name of tag containing cell barcode, or name to store cell barcode under if copying from read name
    cb_regex : str
        Regular expression used to extract cell barcode from read name
    from_tag : bool
        Copy cell barcodes from read tag to read name. If False, copy from read name to tag
    out_format : str
        One of "t" (SAM), "b" (BAM), or "u" (uncompressed BAM) ("t" default)
    """
    assert len(cb_tag) == 2
    
    inputBam = pysam.AlignmentFile(bam)
    outputBam = pysam.AlignmentFile(
        output, "w" + OUT_FORMAT_CONVERSION[out_format], template=inputBam
    )
    readname_barcode = re.compile(cb_regex)
    with closing(inputBam) as inputBam:
        with closing(outputBam) as outputBam:
            for r in inputBam.fetch(until_eof=True):
                tags = r.get_tags()
                if from_tag:
                    cell_barcode, _ = utils.scan_tags(tags, cb=cb_tag)
                    r.query_name = cell_barcode + ":" + r.query_name
                else:
                    re_match = readname_barcode.search(r.query_name)
                    cell_barcode = re_match.group()
                    r.set_tag(cb_tag, cell_barcode, value_type="Z", replace=True)
                outputBam.write(r)



import pysam
import gzip
from multiprocessing import Pool
import functools
import random
import string
import os
from sinto import utils


def _add_read_tags(intervals, bam, sam, output, cb, trim_suffix, mode):
    inputBam = pysam.AlignmentFile(bam, "rb")
    header = inputBam.header.to_dict()
    newhead = dict((k, header[k]) for k in ("HD", "SQ", "RG"))
    ident = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(6)
    )
    outfmt = 'w' if sam else 'wb'
    outputBam = pysam.AlignmentFile(output + ident, outfmt, header=newhead)
    for i in intervals:
        for r in inputBam.fetch(i[0], i[1], i[2]):
            if mode == "tag":
                cell_barcode, _ = utils.scan_tags(r.tags)
            elif mode == "readname":
                cell_barcode = r.qname.split(":")[0]
            else:
                raise Exception("Unknown mode. Use either tag or readname")
            if cell_barcode is not None:
                if trim_suffix:
                    cell_barcode = cell_barcode[:-2]
                if cell_barcode in cb.keys():
                    r.tags += cb[cell_barcode]
            outputBam.write(r)
    outputBam.close()
    inputBam.close()
    return output + ident


def addtags(bam, tagfile, output, sam=False, trim_suffix=True, mode="tag", nproc=1):
    """Add tags to reads from individual cells

    Copies BAM entries to a new file, adding a read tag to cells matching an input table

    Parameters
    ----------
    bam : str
        Path to BAM file.
    tagfile : str
        Tab-delimited file containing cell barcode, read tag to be added, tag information
    output : str
        Name for output BAM file.
    sam : bool, optional
        Output SAM format. Default is BAM format.
    trim_suffix: bool, optional
        Remove trailing 2 characters from cell barcode in bam file (sometimes needed to match 10x barcodes).
    nproc : int, optional
        Number of processors to use. Default is 1.
    mode : str
        Either tag (default) or readname. Some BAM file store the cell barcode in the readname rather than under
        a read tag.

    Raises
    ------
    Exception
        If samtools merge of temporary BAM files fails
    """
    nproc = int(nproc)
    tags = utils.read_cell_barcode_tag_file(tagfile)
    inputBam = pysam.AlignmentFile(bam, "rb")
    intervals = utils.chunk_bam(inputBam, nproc)
    inputBam.close()
    p = Pool(nproc)
    tempfiles = p.map_async(
        functools.partial(
            _add_read_tags,
            bam=bam,
            sam=sam,
            output=output,
            cb=tags,
            trim_suffix=trim_suffix,
            mode=mode,
        ),
        intervals.values(),
    ).get(9999999)
    pysam.merge('-@', str(nproc), '--no-PG', '-h', bam, '-c', output, *tempfiles)
    if os.path.exists(output):
        [os.remove(i) for i in tempfiles]
    else:
        raise Exception("samtools merge failed, temp files not deleted")



import sys
from argparse import ArgumentParser
import pkg_resources
from sinto import cli


version = pkg_resources.require("sinto")[0].version
parser = ArgumentParser(description="Tools for single-cell data processing")
parser.add_argument(
    "-v", "--version", action="version", version="%(prog)s " + str(version)
)
subparsers = parser.add_subparsers(title="Subcommands")

# filterbarcodes
parser_filterbarcodes = subparsers.add_parser(
    "filterbarcodes", description="Filter reads based on input list of cell barcodes"
)
parser_filterbarcodes.add_argument(
    "-b", "--bam", help="Input bam file (must be indexed)", required=True, type=str
)
parser_filterbarcodes.add_argument(
    "-c",
    "--cells",
    help="File or comma-separated list of cell barcodes. Can be gzip compressed",
    required=True,
    type=str,
)
parser_filterbarcodes.add_argument(
    "-t",
    "--trim_suffix",
    help="Remove trail 2 characters from cell barcode in BAM file",
    action="store_true",
    default=False,
)
parser_filterbarcodes.add_argument(
    "-p",
    "--nproc",
    help="Number of processors (default = 1)",
    required=False,
    default=1,
    type=int,
)
parser_filterbarcodes.add_argument(
    "--barcode_regex",
    help="""
    Regular expression used to extract cell barcode from read name. If None (default), extract cell barcode from read tag.
    Use "[^:]*" to match all characters up to the first colon.
    """,
    required=False,
    type=str,
    default=None,
)
parser_filterbarcodes.add_argument(
    "--barcodetag",
    help='Read tag storing cell barcode information (default = "CB")',
    required=False,
    type=str,
    default="CB",
)
parser_filterbarcodes.add_argument(
    "--outdir",
    help='Output file directory',
    required=False,
    type=str,
    default='.',
)
parser_filterbarcodes.add_argument(
    "-s",
    "--sam",
    help="Output sam format (default bam output)",
    required=False,
    action="store_true",
    default=False,
)
parser_filterbarcodes.set_defaults(func=cli.run_filterbarcodes)

# addtags
parser_addtags = subparsers.add_parser(
    "addtags", description="Add read tags to reads from individual cells"
)
parser_addtags.add_argument(
    "-b", "--bam", help="Input bam file (must be indexed)", required=True, type=str
)
parser_addtags.add_argument(
    "-f",
    "--tagfile",
    help="Tab-delimited file containing cell barcode, tag to be added, and tag identity. Can be gzip compressed",
    required=True,
    type=str,
)
parser_addtags.add_argument(
    "-o", "--output", help="Name for output text file", required=True, type=str
)
parser_addtags.add_argument(
    "-t",
    "--trim_suffix",
    help="Remove trail 2 characters from cell barcode in BAM file",
    action="store_true",
    default=False,
)
parser_addtags.add_argument(
    "-s",
    "--sam",
    help="Output sam format (default bam output)",
    required=False,
    action="store_true",
    default=False,
)
parser_addtags.add_argument(
    "-p",
    "--nproc",
    help="Number of processors (default = 1)",
    required=False,
    default=1,
    type=int,
)
parser_addtags.add_argument(
    "-m",
    "--mode",
    help="Either tag (default) or readname. Some BAM file store the cell barcode in the readname rather than under a read tag",
    required=False,
    default="tag",
    type=str,
)
parser_addtags.set_defaults(func=cli.run_addtags)

# tagtorg
parser_tagtorg = subparsers.add_parser(
    "tagtorg",
    description="Append a read tag to the read group ID of each read. "
    "Also appends the read tag to the SM field of the read group.",
)
parser_tagtorg.add_argument(
    "-b",
    "--bam",
    help="Input SAM/BAM file, '-' reads from stdin",
    required=True,
    type=str,
)
parser_tagtorg.add_argument(
    "--tag",
    help="Read tag to extract the value from that is appended to the read group. "
    "Default is 'CB', the tag that is used in 10x sequencing to identify cells.",
    default="CB",
    type=str,
)
parser_tagtorg.add_argument(
    "-f",
    "--tagfile",
    help="List of expected tag values. Reads with tag values that are not in this list are not altered.",
    required=True,
    type=str,
)
parser_tagtorg.add_argument(
    "-o",
    "--output",
    help="Output SAM/BAM file, '-' outputs to stdout (default '-')",
    default="-",
    type=str,
)
parser_tagtorg.add_argument(
    "-O",
    "--outputformat",
    help="Output format. One of 't' (SAM), 'b' (BAM),"
    " or 'u' (uncompressed BAM) ('t' default)",
    default="t",
)


parser_tagtorg.set_defaults(func=cli.run_tagtorg)

# tagtotag
parser_tagtotag = subparsers.add_parser(
    "tagtotag",
    description="Copies BAM entries to a new file while copying a read tag"
                " to another read tag"
                " and optionally deleting the originating tag.",
)
parser_tagtotag.add_argument(
    "-b",
    "--bam",
    help="Input SAM/BAM file, '-' reads from stdin",
    required=True,
    type=str,
)
parser_tagtotag.add_argument(
    "--from",
    help="Read tag to copy from.",
    required=True,
    dest='from_',
    type=str,
)
parser_tagtotag.add_argument(
    "--to",
    help="Read tag to copy to.",
    required=True,
    type=str,
)
parser_tagtotag.add_argument(
    "--delete",
    help="Delete originating tag after copy (i.e. move).",
    action='store_true',
)
parser_tagtotag.add_argument(
    "-o",
    "--output",
    help="Output SAM/BAM file, '-' outputs to stdout (default '-')",
    default="-",
    type=str,
)
parser_tagtotag.add_argument(
    "-O",
    "--outputformat",
    help="Output format. One of 't' (SAM), 'b' (BAM),"
    " or 'u' (uncompressed BAM) ('t' default)",
    default="t",
)


parser_tagtotag.set_defaults(func=cli.run_tagtotag)

# fragments
parser_fragments = subparsers.add_parser(
    "fragments", description="Create ATAC-seq fragment file from BAM file"
)
parser_fragments.add_argument(
    "-b", "--bam", help="Input bam file (must be indexed)", required=True, type=str
)
parser_fragments.add_argument(
    "-f",
    "--fragments",
    help="""
    Name and path for output fragments file. Note that the output is not sorted or compressed.
    To sort the output file use sort -k 1,1 -k2,2n
    """,
    required=True,
    type=str,
)
parser_fragments.add_argument(
    "-m",
    "--min_mapq",
    help="Minimum MAPQ required to retain fragment (default = 30)",
    required=False,
    default=30,
    type=int,
)
parser_fragments.add_argument(
    "-p",
    "--nproc",
    help="Number of processors (default = 1)",
    required=False,
    type=int,
    default=1,
)
parser_fragments.add_argument(
    "-t",
    "--barcodetag",
    help='Read tag storing cell barcode information (default = "CB")',
    required=False,
    type=str,
    default="CB",
)
parser_fragments.add_argument(
    "-c",
    "--cells",
    help="""
    Path to file containing cell barcodes to retain, or a
    comma-separated list of cell barcodes.
    If None (default), use all cell barocodes present in the BAM file.
    """,
    required=False,
    type=str,
    default=None,
)
parser_fragments.add_argument(
    "--barcode_regex",
    help="""
    Regular expression used to extract cell barcode from read name. If None (default), extract cell barcode from read tag.
    Use "[^:]*" to match all characters up to the first colon.
    """,
    required=False,
    type=str,
    default=None,
)
parser_fragments.add_argument(
    "--use_chrom",
    help="""
    Regular expression used to match chromosomes to be included in output.
    Default is "(?i)^chr" to match all chromosomes starting with "chr",
    case insensitive
    """,
    required=False,
    default="(?i)^chr",
    type=str,
)
parser_fragments.add_argument(
    "--max_distance",
    help="""
    Maximum distance between integration sites for the fragment to be retained.
    Allows filtering of implausible fragments that likely result from incorrect
    mapping positions. Default is 5000 bp.
    """,
    required=False,
    default=5000,
    type=int,
)
parser_fragments.add_argument(
    "--min_distance",
    help="""
    Minimum distance between integration sites for the fragment to be retained.
    Allows filtering of implausible fragments that likely result from incorrect
    mapping positions. Default is 10 bp.
    """,
    required=False,
    default=10,
    type=int,
)
parser_fragments.add_argument(
    "--chunksize",
    help="""
    Number of BAM file entries to iterate over before collapsing the fragments and writing
    to disk. Higher chunksize will use more memory but will be faster.
    """,
    required=False,
    default=500000,
    type=int,
)
parser_fragments.add_argument(
    "--shift_plus",
    help="""
    Number of bases to shift Tn5 insertion position by on the forward strand.
    """,
    required=False,
    default=4,
    type=int,
)
parser_fragments.add_argument(
    "--shift_minus",
    help="""
    Number of bases to shift Tn5 insertion position by on the reverse strand.
    """,
    required=False,
    default=-5,
    type=int,
)
parser_fragments.add_argument(
    "--collapse_within",
    help="""
    Take cell barcode into account when collapsing duplicate fragments. Setting this
    flag means that fragments with the same coordinates can be identified provided 
    they originate from a different cell barcode.
    """,
    action="store_true",
    default=False,
)
parser_fragments.set_defaults(func=cli.run_fragments)

# barcodes
parser_barcode = subparsers.add_parser(
    "barcode", description="Add cell barcode sequences to read names in FASTQ file."
)
parser_barcode.add_argument(
    "--barcode_fastq", help="FASTQ file containing cell barcode sequences", required=True, type=str
)
parser_barcode.add_argument(
    "--read1", help="FASTQ file containing read 1", required=True, type=str
)
parser_barcode.add_argument(
    "--read2", help="FASTQ file containing read 2", required=False, type=str, default=None
)
parser_barcode.add_argument(
    "-b", "--bases", help="Number of bases to extract from barcode-containing FASTQ", required=True, type=int
)
parser_barcode.add_argument(
    "--prefix", help="Prefix to add to cell barcodes", required=False, type=str, default=""
)
parser_barcode.add_argument(
    "--suffix", help="Suffix to add to cell barcodes", required=False, type=str, default=""
)
parser_barcode.set_defaults(func=cli.run_barcode)

# tagtoname
parser_tagtoname = subparsers.add_parser(
    "tagtoname",
    description="""
    Copy cell barcode sequences from tag to read names. Cell barcodes will be added
    as a readname prefix, followed by ":"
    """
)
parser_tagtoname.add_argument(
    "-b",
    "--bam",
    help="Input SAM/BAM file, '-' reads from stdin",
    required=True,
    type=str,
)
parser_tagtoname.add_argument(
    "-o",
    "--output",
    help="Output SAM/BAM file, '-' outputs to stdout (default '-')",
    default="-",
    type=str,
)
parser_tagtoname.add_argument(
    "-O",
    "--outputformat",
    help="Output format. One of 't' (SAM), 'b' (BAM),"
    " or 'u' (uncompressed BAM) ('t' default)",
    default="t",
)
parser_tagtoname.add_argument(
    "--tag",
    help="Read tag to copy from.",
    required=False,
    default="CB",
    type=str,
)
parser_tagtoname.set_defaults(func=cli.run_tagtoname)

# nametotag
parser_nametotag = subparsers.add_parser(
    "nametotag", description="Copy cell barcode sequences from read name to read tag"
)
parser_nametotag.add_argument(
    "-b",
    "--bam",
    help="Input SAM/BAM file, '-' reads from stdin",
    required=True,
    type=str,
)
parser_nametotag.add_argument(
    "-o",
    "--output",
    help="Output SAM/BAM file, '-' outputs to stdout (default '-')",
    default="-",
    type=str,
)
parser_nametotag.add_argument(
    "-O",
    "--outputformat",
    help="Output format. One of 't' (SAM), 'b' (BAM),"
    " or 'u' (uncompressed BAM) ('t' default)",
    default="t",
)
parser_nametotag.add_argument(
    "--barcode_regex",
    help="""
    Regular expression used to extract cell barcode from read name.
    Default ("[^:]*") matches all characters up to the first colon.
    """,
    required=False,
    type=str,
    default="[^:]*",
)
parser_nametotag.add_argument(
    "--tag",
    help="Read tag to copy to.",
    required=False,
    default="CB",
    type=str,
)
parser_nametotag.set_defaults(func=cli.run_nametotag)


def main():
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    else:
        options = parser.parse_args()
        sys.setrecursionlimit(200000)
        options.func(options)

if __name__ == "__main__":
    main()
